# calibcant - tools for thermally calibrating AFM cantilevers
#
# Copyright (C) 2008-2012 W. Trevor King <wking@drexel.edu>
#
# This file is part of calibcant.
#
# calibcant is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# calibcant is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# calibcant.  If not, see <http://www.gnu.org/licenses/>.

"""Temperature analysis.

Separate the more general `T_analyze()` from the other `T_*()`
functions in calibcant.

The relevant physical quantities are:

* `T` Temperature at which thermal vibration measurements were acquired

>>> import os
>>> import tempfile
>>> import numpy
>>> from .config import TemperatureConfig
>>> from h5config.storage.hdf5 import pprint_HDF5, HDF5_Storage

>>> fd,filename = tempfile.mkstemp(suffix='.h5', prefix='calibcant-')
>>> os.close(fd)

>>> temperature_config = TemperatureConfig(storage=HDF5_Storage(
...         filename=filename, group='/T/config/'))

>>> raw_T = numpy.array([22, 23.5, 24])
>>> processed_T = T_analyze(raw_T, temperature_config)
>>> T_plot(raw_T=raw_T, processed_T=processed_T)
>>> T_save(filename=filename, group='/T/', raw_T=raw_T,
...     temperature_config=temperature_config, processed_T=processed_T)

>>> pprint_HDF5(filename)  # doctest: +REPORT_UDIFF
/
  /T
    /T/config
      <HDF5 dataset "default": shape (), type "|b1">
        True
      <HDF5 dataset "units": shape (), type "|S7">
        Celsius
    <HDF5 dataset "processed": shape (3,), type "<f8">
      [ 295.15  296.65  297.15]
    <HDF5 dataset "raw": shape (3,), type "<f8">
      [ 22.   23.5  24. ]

>>> raw_T,temperature_config,processed_T = T_load(
...     filename=filename, group='/T/')
>>> print temperature_config.dump()
units: Celsius
default: yes
>>> raw_T
array([ 22. ,  23.5,  24. ])
>>> type(raw_T)
<type 'numpy.ndarray'>
>>> processed_T
array([ 295.15,  296.65,  297.15])

>>> os.remove(filename)
"""

import h5py as _h5py
from scipy.constants import C2K as _C2K

try:
    import matplotlib as _matplotlib
    import matplotlib.pyplot as _matplotlib_pyplot
    import time as _time  # for timestamping lines on plots
except (ImportError, RuntimeError), e:
    _matplotlib = None
    _matplotlib_import_error = e

from h5config.storage.hdf5 import HDF5_Storage as _HDF5_Storage
from h5config.storage.hdf5 import h5_create_group as _h5_create_group

from . import LOG as _LOG
from . import package_config as _package_config
from .config import Celsius as _Celsius
from .config import Kelvin as _Kelvin
from .config import TemperatureConfig as _TemperatureConfig


def T_analyze(T, temperature_config):
    """Convert measured temperature to Kelvin.

    `T` should be a numpy ndarray or scalar.  `temperature_config`
    should be a `config._TemperatureConfig` instance.
    """
    if temperature_config['units'] == _Celsius:
        return _C2K(T)
    else:
        return T

def T_save(filename, group='/', raw_T=None, temperature_config=None,
           processed_T=None):
    with _h5py.File(filename, 'a') as f:
        cwg = _h5_create_group(f, group)
        if raw_T is not None:
            try:
                del cwg['raw']
            except KeyError:
                pass
            cwg['raw'] = raw_T
        if temperature_config is not None:
            config_cwg = _h5_create_group(cwg, 'config')
            storage = _HDF5_Storage()
            storage.save(config=temperature_config, group=config_cwg)
        if processed_T is not None:
            try:
                del cwg['processed']
            except KeyError:
                pass
            cwg['processed'] = processed_T

def T_load(filename, group='/'):
    assert group.endswith('/')
    raw_T = processed_T = None
    with _h5py.File(filename, 'a') as f:
        try:
            raw_T = f[group+'raw'][...]
        except KeyError:
            pass
        temperature_config = _TemperatureConfig(storage=_HDF5_Storage(
                filename=filename, group=group+'config/'))
        try:
            processed_T = f[group+'processed'][...]
        except KeyError:
            pass
    temperature_config.load()
    return (raw_T, temperature_config, processed_T)

def T_plot(raw_T=None, processed_T=None):
    if not _matplotlib:
        raise _matplotlib_import_error
    figure = _matplotlib_pyplot.figure()
    timestamp = _time.strftime('%H%M%S')
    if raw_T is None:
        if processed_T is None:
            return  # nothing to plot
        axes1 = None
        axes2 = figure.add_subplot(1, 1, 1)
    elif processed_T is None:
        axes1 = figure.add_subplot(1, 1, 1)
        axes2 = None
    else:
        axes1 = figure.add_subplot(2, 1, 1)
        axes2 = figure.add_subplot(2, 1, 2)
    if axes1:
        axes1.set_title('Raw Temperatures %s' % timestamp)
        axes1.plot(raw_T, label='raw')
    if axes2:
        axes2.set_title('Processed Temperatures %s' % timestamp)
        axes2.plot(processed_T, label='processed')
    if hasattr(figure, 'show'):
        figure.show()
