# calibcant - tools for thermally calibrating AFM cantilevers
#
# Copyright (C) 2011-2012 W. Trevor King <wking@drexel.edu>
#
# This file is part of calibcant.
#
# calibcant is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# calibcant is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# calibcant.  If not, see <http://www.gnu.org/licenses/>.

"""Acquire, save, and load cantilever thermal vibration data.

For measuring the cantilever's spring constant.

The relevent physical quantity is:
  Vphoto   The photodiode vertical deflection voltage (what we measure)
"""

from numpy import zeros as _zeros
from FFT_tools import ceil_pow_of_two as _ceil_pow_of_two
from pycomedi.constant import TRIG_SRC as _TRIG_SRC
from pycomedi.utility import inttrig_insn as _inttrig_insn
from pycomedi.utility import Reader as _Reader

from . import LOG as _LOG
from .vib_analyze import vib_analyze as _vib_analyze
from .vib_analyze import vib_save as _vib_save


def vib_acquire(piezo, vibration_config):
    """Record thermal vibration data for `piezo` at its current position.

    Inputs:
      piezo             a pypiezo.afm.AFMPiezo instance
      vibration_config  a .config._VibrationConfig instance
    """
    _LOG.debug('prepare vibration aquisition command')

    # round up to the nearest power of two, for efficient FFT-ing
    n_samps = _ceil_pow_of_two(
        vibration_config['sample-time']*vibration_config['frequency'])
    time = n_samps / vibration_config['frequency']
    scan_period_ns = int(1e9 / vibration_config['frequency'])

    input_channel = piezo.input_channel_by_name('deflection')
    channel = input_channel.channel

    channels = [channel]

    dtype = piezo.deflection_dtype()
    data = _zeros((n_samps, len(channels)), dtype=dtype)

    cmd = channel.subdevice.get_cmd_generic_timed(
        len(channels), scan_period_ns)
    cmd.start_src = _TRIG_SRC.int
    cmd.start_arg = 0
    cmd.stop_src = _TRIG_SRC.count
    cmd.stop_arg = n_samps
    cmd.chanlist = channels
    channel.subdevice.cmd = cmd
    for i in range(3):
        rc = channel.subdevice.command_test()
        if rc == None: break
        _LOG.debug('command test %d: %s' % (i,rc))

    _LOG.info('get %g seconds of vibration data at %g Hz'
              % (vibration_config['sample-time'],
                 vibration_config['frequency']))
    channel.subdevice.command()
    reader = _Reader(channel.subdevice, data)
    reader.start()
    channel.subdevice.device.do_insn(_inttrig_insn(channel.subdevice))
    reader.join()
    data = data.reshape((data.size,))
    return data

def vib(piezo, vibration_config, filename, group='/'):
    """Wrapper around vib_acquire(), vib_analyze(), vib_save().

    >>> import os
    >>> import tempfile
    >>> from h5config.storage.hdf5 import pprint_HDF5
    >>> from pycomedi.device import Device
    >>> from pycomedi.subdevice import StreamingSubdevice
    >>> from pycomedi.channel import AnalogChannel
    >>> from pycomedi.constant import AREF, SUBDEVICE_TYPE, UNIT
    >>> from pypiezo.afm import AFMPiezo
    >>> from pypiezo.base import InputChannel
    >>> from pypiezo.config import ChannelConfig
    >>> from .config import VibrationConfig

    Setup an `AFMPiezo` instance.

    >>> fd,filename = tempfile.mkstemp(suffix='.h5', prefix='calibcant-')
    >>> os.close(fd)

    >>> d = Device('/dev/comedi0')
    >>> d.open()

    >>> s_in = d.find_subdevice_by_type(SUBDEVICE_TYPE.ai,
    ...     factory=StreamingSubdevice)

    >>> channel = s_in.channel(0, factory=AnalogChannel, aref=AREF.diff)
    >>> channel.range = channel.find_range(
    ...     unit=UNIT.volt, min=-10, max=10)
    >>> channel_config = ChannelConfig()
    >>> channel_config['name'] = 'deflection'

    >>> c = InputChannel(config=channel_config, channel=channel)
    >>> c.setup_config()

    >>> piezo = AFMPiezo(axes=[], inputs=[c])

    Test a vibration:

    >>> vibration_config = VibrationConfig()
    >>> vib(piezo, vibration_config, filename, group='/vibration')
    TODO: replace skipped example data with real-world values
    >>> pprint_HDF5(filename)  # doctest: +ELLIPSIS, +REPORT_UDIFF
    /
      /vibration
        /vibration/config
          /vibration/config/deflection
            <HDF5 dataset "channel": shape (), type "<i4">
              0
            <HDF5 dataset "conversion-coefficients": shape (2,), type "<f8">
              [ -1.00000000e+01   3.05180438e-04]
            <HDF5 dataset "conversion-origin": shape (), type "<f8">
              0.0
            <HDF5 dataset "device": shape (), type "|S12">
              /dev/comedi0
            <HDF5 dataset "inverse-conversion-coefficients": shape (2,), type "<f8">
              [    0.    3276.75]
            <HDF5 dataset "inverse-conversion-origin": shape (), type "<f8">
              -10.0
            <HDF5 dataset "maxdata": shape (), type "<i8">
              65535
            <HDF5 dataset "name": shape (), type "|S10">
              deflection
            <HDF5 dataset "range": shape (), type "<i4">
              0
            <HDF5 dataset "subdevice": shape (), type "<i4">
              0
          /vibration/config/vibration
            <HDF5 dataset "chunk-size": shape (), type "<i4">
              2048
            <HDF5 dataset "frequency": shape (), type "<f8">
              50000.0
            <HDF5 dataset "maximum-fit-frequency": shape (), type "<f8">
              25000.0
            <HDF5 dataset "minimum-fit-frequency": shape (), type "<f8">
              500.0
            <HDF5 dataset "model": shape (), type "|S12">
              Breit-Wigner
            <HDF5 dataset "overlap": shape (), type "|b1">
              False
            <HDF5 dataset "sample-time": shape (), type "<i4">
              1
            <HDF5 dataset "window": shape (), type "|S4">
              Hann
        <HDF5 dataset "processed": shape (), type "<f8">
          ...
        /vibration/raw
          <HDF5 dataset "deflection": shape (65536,), type "<u2">
            [...]

    Close the Comedi device.

    >>> d.close()

    Cleanup our temporary config file.

    >>> os.remove(filename)
    """
    deflection_input_channel = piezo.input_channel_by_name('deflection')
    
    deflection_channel_config = deflection_input_channel.config

    deflection = vib_acquire(piezo, vibration_config)
    variance = _vib_analyze(
        deflection, vibration_config, deflection_channel_config)
    _vib_save(
        filename, group, raw_vibration=deflection,
        vibration_config=vibration_config,
        deflection_channel_config=deflection_channel_config,
        processed_vibration=variance)
    return variance
